package pro.requena.tfm;

import java.awt.Color;
import java.awt.Container;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.nio.file.StandardOpenOption;
import java.util.ArrayList;
import java.util.List;

import javax.sound.sampled.AudioFormat;
import javax.sound.sampled.AudioSystem;
import javax.sound.sampled.LineUnavailableException;
import javax.sound.sampled.Mixer;
import javax.sound.sampled.Mixer.Info;
import javax.sound.sampled.TargetDataLine;
import javax.swing.GroupLayout;
import javax.swing.JButton;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JSlider;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

public class App extends JPanel {
	private static final long serialVersionUID = 1L;
	static AudioFormat audioFormat = new AudioFormat(44100, 16, 1, true, false);
	private static final int sampleRate = 2048;
	static double[] data = new double[sampleRate];
	static double[] espectre = new double[sampleRate];
	static double[] cepstrum = new double[sampleRate];
	static int threshold = 5;
	private static boolean isRecording = false;
	// Inicialització MFCC.
	private static int mfccBands = 12;
	private static MFCC mfcc = new MFCC(sampleRate, 44100, mfccBands);

	private static Container myContainer;
	private static JFrame myWindow;

	// Switches
	private static boolean activarAutoDeteccio = true;
	private static boolean activarNormalitzacioEspectre = true;
	private static boolean activarDebugValorsMFCC = false;
	

	// Emmagatzematge de l'MFCC.
	private static List<double[]> mfccArray = new ArrayList<double[]>();

	public static void main(String[] args) {
		// Crear finestra
		myWindow = new JFrame("TFM");
		myContainer = myWindow.getContentPane();
		GroupLayout groupLayout = new GroupLayout(myContainer);
		groupLayout.setAutoCreateContainerGaps(true);
		groupLayout.setAutoCreateGaps(true);
		myContainer.setLayout(groupLayout);

		// Panell de botons
		JPanel panBotons = new JPanel();
		JButton cmdGravar = new JButton("Gravar");
		cmdGravar.addActionListener(new GravarAL());
		JSlider thresoldSlider = new JSlider(0, 50);
		thresoldSlider.setValue(5);
		thresoldSlider.setMajorTickSpacing(10);
		thresoldSlider.setMinorTickSpacing(1);
		thresoldSlider.setPaintLabels(true);
		thresoldSlider.setPaintTicks(true);
		thresoldSlider.addChangeListener(new CanviThresold());
		panBotons.add(cmdGravar);
		panBotons.add(thresoldSlider);
		panBotons.setSize(1000, 50);

		// Panell de dades
		JPanel panDades = new JPanel();
		JLabel lblMitjana = new JLabel("Sense dades");
		panDades.add(lblMitjana);
		JLabel lblDesviacio = new JLabel("Sense dades");
		panDades.add(lblDesviacio);
		panDades.setSize(1000, 50);

		// Gràfic
		Grafic myGrafic = new Grafic();

		// Gràfic per l'espectre
		Espectre myEspectre = new Espectre();

		// Gràfic pel cepstrum
		Cepstrum myCepstrum = new Cepstrum();

		groupLayout.setVerticalGroup(
			groupLayout.createSequentialGroup()
				.addComponent(panBotons, 50, 50, 50)
				.addComponent(panDades, 50, 50, 50)
				.addComponent(myGrafic, 256, 256, 256)
				.addComponent(myEspectre, 256, 256, 256)
				.addComponent(myCepstrum, 256, 256, 256)
		);
		groupLayout.setHorizontalGroup(
			groupLayout.createSequentialGroup()
				.addGroup(groupLayout.createParallelGroup()
					.addComponent(panBotons)
					.addComponent(panDades)
					.addComponent(myGrafic)
					.addComponent(myEspectre)
					.addComponent(myCepstrum)
				)
		);
		

		myWindow.setSize(1200, 1000);
		myWindow.setVisible(true);
		myWindow.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);

		Mixer mixer = null;
		for (Info info : AudioSystem.getMixerInfo()) {
			if (info.getName().equals("Micrófono (Dispositivo de High "))
			{
				mixer = AudioSystem.getMixer(info);
			}
		}

		TargetDataLine line;

		// Obtain and open the line.
		try {
			line = (TargetDataLine) AudioSystem.getLine(mixer.getTargetLineInfo()[0]);
			line.open(audioFormat);

			// Principal.
			long comptadorSilenci = 0L;
			boolean stopped = false;
			int numBytesRead;
			
			line.start();

			while (!stopped) {
				byte[] readBytes = new byte[sampleRate*2];
				numBytesRead = line.read(readBytes, 0, data.length);

				// Transformar els bytes llegits en doubles.
				obtenirDades(readBytes);

				// Anul·lar el soroll de fons.
				for (int i=0; i<numBytesRead; i++) {
					if (Math.abs(data[i]*256) < threshold) {
						data[i] = 0;
					}
				}

				// Actualitzar controls.
				// Calcular la mitjana.
				double calcul = 0L;
				int mitjana = 0;
				for (int i=0; i<numBytesRead; i++) {
					calcul = calcul + data[i];
				}
				mitjana = (int) (calcul / numBytesRead);

				// Calcular la desviació estàndard.
				double calculDouble = 0F;
				Float desviacio = 0F;
				for (int i=0; i<numBytesRead; i++) {
					calculDouble = calculDouble + Math.pow(data[i] - mitjana, 2);
				}
				desviacio = (float) (calculDouble / numBytesRead);

				if (!isRecording) {
					if (desviacio > 0.0) {
						if (activarAutoDeteccio)
							cmdGravar.doClick();
					}
				}
				if (isRecording) {
					if (desviacio == 0.0) {
						if (comptadorSilenci > 0) {
							comptadorSilenci = 0;
							if (activarAutoDeteccio)
								cmdGravar.doClick();
						}
						else {
							comptadorSilenci++;
						}
					}
				}

				if (isRecording) {
					// Mostrar dades estadístiques calculades anteriorment.
					lblMitjana.setText("Mitjana: " + mitjana);
					lblDesviacio.setText("σ: " + desviacio);

					// Guardar fitxer.
					Files.write(Paths.get("recording.kav"), readBytes, StandardOpenOption.APPEND);

					// Repintar gràfic.
					myGrafic.repaint();

					// Aplicar FFT a la senyal per obtenir l'espectre.
					espectre = Fft.fft(data, new double[sampleRate], true);

					// Normalització de l'espectre perquè el seu valor màxim sigui 1048633.4400395132.
					if (activarNormalitzacioEspectre) {
						normalitzarEspectre();
					}

					// Aplicar IDFT a l'espectre per obtenir el cepstrum.
					// Primer cal obtenir el log en base 10 del valor absolut del vector.
					double[] temp = new double[sampleRate];
					for (int i=0;i<temp.length;i++) {
						temp[i] = Math.log10(Math.abs(espectre[i]));
					}
					cepstrum = Fft.fft(temp, new double[sampleRate], false);


					// Aplicar MFCC a l'espectre.
					double[] mel = mfcc.doMFCC(espectre);

					// Emmagatzemar l'MFCC.
					mfccArray.add(mel);

					// Debug: Printar la mitjana dels exponents.
					double[] calculsMfcc = new double[mfccBands];
					for (int i=0;i<calculsMfcc.length;i++)
						calculsMfcc[i] = 0;

					for (double[] dada : mfccArray) {
						for (int i=0;i<dada.length;i++)
							calculsMfcc[i] = calculsMfcc[i] + dada[i] / mfccArray.size();
					}

					// Generar una clau a partir dels rangs de l'MFCC.
					Integer[] seed = new Integer[mfccBands];
					Double[] factorPosicio = new Double[]{20.0, 3.0, 1.0, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5}; 
					for (int i=0;i<mfccBands;i++) {
						// Càrrega de l'array "seed" amb els valors ponderats de les bandes de l'MFCC.
						seed[i]=(int) (calculsMfcc[i] / factorPosicio[i]);
					}
					debugValorsSeed(seed);

					// Veure valors de depuració
					if (activarDebugValorsMFCC) {
						debugValorsMFCC(calculsMfcc);
					}


					// Sobrescriure el gràfic del Cepstrum.
					for(int i=0;i<cepstrum.length;i++)
						cepstrum[i]=0;
					for(int i=0;i<mfccBands * 50;i++)		// Bandes d'1 a 13 - És on es troben les freq de la veu.
						cepstrum[i] = mel[i / 50] * 100;


					// Conversió d'FFT a escala logarítmica (dB)
					for (int i=0;i<espectre.length;i++) {
						espectre[i]=20*Math.log10(espectre[i] + 1);	// Aplico un valor de referència d'1.
					}

					// Refrescar gràfics.
					myEspectre.repaint();
					myCepstrum.repaint();
				}
			}
		

			line.drain();
			line.close();
		} catch (LineUnavailableException ex) {
			System.out.println("Error: Line no disponible");
		}
		 catch (IOException ex) {
			 System.out.println("Error: No s'ha pogut guardar el fitxer 'recording.kav'");
		 }

		System.out.println("Sortida.");
	}

	private static void debugValorsMFCC(double[] calculsMfcc) {
		for (int i=0;i<calculsMfcc.length;i++) {
			System.out.print(calculsMfcc[i] + "\t");
		}
		System.out.println();

		int thresoldReconeixement = 20;		// Diferència proporcional a 10 punts sobre 100.
		Double[] vA = new Double[]{229.0360731976437, 30.14054389931937, 11.572529398224866, 2.6934956964384944, -1.023615724035677, -3.5718527256759844, -3.356359319519307, -1.8255612502606997, 0.0754356192393009, 1.5977346203625005, 1.645316566556175, 1.3250469098928361};
		Double[] vE = new Double[]{267.1317379505319, 23.005902737345526, 9.611775511962847, 3.7497022630063235, 4.497959569349822, 4.627485118912615, 2.9020800636231767, 0.8099977671052998, -0.6531367005498515, -0.41579736354497926, -0.008883042530070127, 0.7939510717188651};
		Double[] vI = new Double[]{256.1045357622962, 22.182724160151693, 12.253039272817436, 5.322680667736602, 5.378486627723827, 4.784827138452641, 2.6364072574279893, 0.995798868181868, 1.094087636016987, 1.5574590009310607, 1.345148722139366, 0.6335920616395914};

		boolean reconeixement = true;
		for (int i=1;i<mfccBands;i++) {
			if (Math.abs(calculsMfcc[i] - vA[i]) > (thresoldReconeixement / calculsMfcc[i] * 100))
//						if (Math.abs(calculsMfcc[i] - vA[i]) > thresoldReconeixement)
				reconeixement = false;
		}
		if (reconeixement)
			System.out.println("S'ha reconegut la clau 'A'.");

		reconeixement = true;
		for (int i=1;i<mfccBands;i++) {
			if (Math.abs(calculsMfcc[i] - vE[i]) > (thresoldReconeixement / calculsMfcc[i] * 100))
				reconeixement = false;
		}
		if (reconeixement)
			System.out.println("S'ha reconegut la clau 'E'.");

		reconeixement = true;
		for (int i=1;i<mfccBands;i++) {
			if (Math.abs(calculsMfcc[i] - vI[i]) > (thresoldReconeixement / calculsMfcc[i] * 100))
				reconeixement = false;
		}
		if (reconeixement)
			System.out.println("S'ha reconegut la clau 'I'.");
	}

	private static void debugValorsSeed(Integer[] seed) {
		
		// Calcular diferencial entre les components del vector seed[].
		Integer[] diferencial = new Integer[seed.length-1];
		for(int i=0;i<seed.length-1;i++) {
			diferencial[i] = seed[i+1] - seed[i];
			System.out.print(diferencial[i] + "\t");
		}
		System.out.println();
		
		Integer[] vA = new Integer[]{0, 0, -15, -10, 3, 3, 0, 3, 3, 0, -5};
		Integer[] vE = new Integer[]{-2, -2, -5, 10, 5, -15, -15, -5, 5, 5, 0};

		int threshold = 5;
		boolean reconeixement = true;
		for(int i=0;i<diferencial.length-1;i++) {
			if(Math.abs(vA[i] - diferencial[i]) > threshold) {
				reconeixement = false;
			}
		}
		if (reconeixement) {
			System.out.println("Reconeguda: A.");
		}


		reconeixement = true;
		for(int i=0;i<diferencial.length-1;i++) {
			if(Math.abs(vE[i] - diferencial[i]) > threshold) {
				reconeixement = false;
			}
		}
		if (reconeixement) {
			System.out.println("Reconeguda: E.");
		}


		
//		System.out.println("Clau generada: " + generatedKey);
		
		
	}

	private static void normalitzarEspectre() {
		// 1. Cercar el valor mínim/màxim de l'espectre.
		double min=999.0, max=-1.0;
		int minPos=-1, maxPos=-1;
		for (int i=0;i<espectre.length;i++) {
			if (espectre[i]>max) {
				max = espectre[i];
				maxPos = i;
			}
			if (espectre[i]<min) {
				min = espectre[i];
				minPos = i;
			}
		}

		// 2. Obtenir el factor de normalització
		double factor = 1048633.4400395132 / espectre[maxPos];

		// 3. Aplicar el factor de normalització a tot l'espectre.
		for (int i=0;i<espectre.length;i++) {
			espectre[i] = espectre[i] * factor;
		}
	}

	static class GravarAL implements ActionListener {
		public void actionPerformed(ActionEvent arg0) {
			isRecording = !isRecording;

			// Inici de gravació a fitxer.
			if (isRecording) {
				mfccArray.clear();

				((JButton) arg0.getSource()).setText("Aturar");
				try {
					Files.deleteIfExists(Paths.get("recording.kav"));
					Files.createFile(Paths.get("recording.kav"));
				} catch (IOException e) {
					e.printStackTrace();
				}
			}
			else {
				// Gravar coeficients MFCC a CSV.
				CoeficientsMFCC coeficientsMfcc = new CoeficientsMFCC();
				coeficientsMfcc.gravarCSV(mfccArray);

				((JButton) arg0.getSource()).setText("Gravar");
			}
		}
	}

	/**
	 * Aquesta funció actualitza la variable data[] amb els doubles corresponents als bytes d'entrada.
	 * @param byteArray
	 */
    public static void obtenirDades(byte[] byteArray) {  
        assert byteArray.length == 2 * data.length;  
        for (int i = 0; i < data.length; i++) {  
        	data[i] = (short) (((0xFF & byteArray[2 * i + 1]) << 8) | (0xFF & byteArray[2 * i]));  
        	data[i] /= Short.MAX_VALUE;  
        }  
    }  
	
	static class CanviThresold implements ChangeListener {
		public void stateChanged(ChangeEvent arg0) {
			threshold = ((JSlider) arg0.getSource()).getValue();			
		}
	}

	static class Grafic extends JPanel {
		private static final long serialVersionUID = 1L;

		@Override
		public void paint(Graphics g) {
			Graphics2D g2d = (Graphics2D) g;
			g2d.setColor(Color.WHITE);
			g2d.fillRect(0, 0, 2000, 256);

			g2d.setColor(Color.RED);
			int i;
			for (i=0;i<sampleRate;i++) {
				int valor = (int) (data[i]*256);
				g2d.drawLine(i, 128, i, 128 + valor);
			}
		}
	}

	
	static class Espectre extends JPanel {
		private static final long serialVersionUID = 1L;

		@Override
		public void paint(Graphics g) {
			Graphics2D g2d = (Graphics2D) g;
			g2d.setColor(Color.WHITE);
			g2d.fillRect(0, 0, 2000, 256);

			g2d.setColor(Color.RED);
			for (int i=0;i<sampleRate/2;i++) {
				// D'escala dB [0..108.37] a rang 0..256
				int valor = (int) (espectre[i]/108.37*256);

				// Colorejar...
				if (valor <32) {
					g2d.setColor(Color.GREEN);
				}
				else if (valor >= 32 && valor <64) {
					g2d.setColor(Color.YELLOW);
				}
				else if (valor >=64 && valor<96) {
					g2d.setColor(Color.ORANGE);
				}
				else if (valor >=96 && valor<160) {
					g2d.setColor(Color.RED);
				}
				else if (valor >=224) {
					g2d.setColor(Color.MAGENTA);
				}
				g2d.drawLine(i, 256, i, 256-valor);
			}
		}
	}

	static class Cepstrum extends JPanel {
		private static final long serialVersionUID = 1L;

		@Override
		public void paint(Graphics g) {
			Graphics2D g2d = (Graphics2D) g;
			g2d.setColor(Color.WHITE);
			g2d.fillRect(0, 0, 2000, 256);

			g2d.setColor(Color.RED);
			for (int i=0;i<sampleRate/2;i++) {
				// El rang és de 0 a 6*10^7. Normalitzar.
				int valor = (int) (cepstrum[i]/100);

				// Colorejar...
				if (valor <32) {
					g2d.setColor(Color.GREEN);
				}
				else if (valor >= 32 && valor <64) {
					g2d.setColor(Color.YELLOW);
				}
				else if (valor >=64 && valor<96) {
					g2d.setColor(Color.ORANGE);
				}
				else if (valor >=96 && valor<160) {
					g2d.setColor(Color.RED);
				}
				else if (valor >=224) {
					g2d.setColor(Color.MAGENTA);
				}
				g2d.drawLine(i, 256, i, 256-valor);
			}
		}
	}
}
